import os
import uuid
from pathlib import Path

from retrying import retry

from abejacli.bucket.process_file_job import (
    FINISH_REPORT,
    INITIALIZE_REPORT,
    PROGRESS_REPORT,
    RAISE_ERROR,
    SKIP_REPORT
)
from abejacli.config import (
    HTTP_READ_CHUNK_SIZE,
    ORGANIZATION_ENDPOINT,
    PLATFORM_REQUEST_TIMEOUT_SECONDS
)
from abejacli.session import generate_retry_session, generate_user_session


def _get_default_file_path(download_dir: str, file_name: str):
    if file_name.startswith('/'):
        file_name = file_name[1:]
    return os.path.join(download_dir, file_name)


def download_job(bucket_id, file_info, report_queue, options):
    """
    download files until consuming all items in file queue

    :param bucket_id: bucket identifier
    :param file_info: information of file to download
    :param report_queue: queue to report progress for each file
    :param options: option including target download directory
    :return:
    """
    try:
        _download_job_proecess(bucket_id, file_info, report_queue, options)
    except:
        # pass to keep the thread running
        pass


DOWNLOAD_RETRY_ATTEMPT_NUMBER = 3


@retry(stop_max_attempt_number=DOWNLOAD_RETRY_ATTEMPT_NUMBER)
def _download_job_proecess(bucket_id, file_info, report_queue, options):
    """
    download files until consuming all items in file queue

    :param bucket_id: bucket identifier
    :param file_info: information of file to download
    :param report_queue: queue to report progress for each file
    :param options: option including target download directory
    :return:
    """

    publisher_id = uuid.uuid4().hex
    download_dir = options.get('download_dir', '.')

    file_id = file_info.get('file_id')
    download_uri = file_info.get('download_uri')
    file_name = file_id
    download_path = None
    is_downloading = True
    try:
        # download file content
        with generate_retry_session() as session:
            download_stream_res = session.get(
                download_uri, stream=True, timeout=PLATFORM_REQUEST_TIMEOUT_SECONDS)
        # update pre-signed url if expired
        if download_stream_res.status_code == 403:
            url = "{}/buckets/{}/files/{}".format(
                ORGANIZATION_ENDPOINT, bucket_id, file_id)
            with generate_user_session() as user_session:
                res = user_session.get(
                    url, timeout=PLATFORM_REQUEST_TIMEOUT_SECONDS)
            res.raise_for_status()
            file_info = res.json()
            download_uri = file_info.get('download_uri')
            with generate_retry_session() as session:
                download_stream_res = session.get(
                    download_uri, stream=True, timeout=PLATFORM_REQUEST_TIMEOUT_SECONDS)
        download_stream_res.raise_for_status()
        total_size = int(download_stream_res.headers.get('content-length', 0))

        initialize_options = {
            'file_name': file_name,
            'total': total_size,
        }
        report_queue.put(
            (INITIALIZE_REPORT, publisher_id, 0, initialize_options))
        download_path = _get_default_file_path(download_dir, file_name)
        result_options = {
            'source': file_id,
            'destination': download_path,
        }
        if Path(download_path).exists():
            report_queue.put((SKIP_REPORT, publisher_id, 0, result_options))
            return
        else:
            Path(download_path).parent.mkdir(parents=True, exist_ok=True)
        is_downloading = True
        with open(download_path, 'wb') as f:
            for chunk in download_stream_res.iter_content(chunk_size=HTTP_READ_CHUNK_SIZE):
                # update tqdm progress bar with chunk data size
                report_queue.put(
                    (PROGRESS_REPORT, publisher_id, len(chunk), None))
                f.write(chunk)

        report_queue.put((FINISH_REPORT, publisher_id, 0, result_options))
    except Exception:
        if is_downloading and download_path is not None and os.path.exists(download_path):
            os.remove(download_path)
        options = {
            'source': file_id,
            'error': 'Failed to download {} of bucket_id {}'.format(
                file_id, bucket_id)
        }
        report_queue.put((RAISE_ERROR, publisher_id, 0, options))
        raise
